//---------------------------------------------------------------------
//
// NAME: GetReply.c
//
// TYPE: Win32/Win64 console mode.
//
// DATE: May 10, 2021.
//
// PURPOSE:
//
//   Run this program after email has been sent using the SendMail example program.
//
//   This program connects to your POP3 server and reads reply messages:
//
//     (1) From recipient's SMTP server when the email previously sent to the client was undeliverable.
//     (2) From the recipient when he replies - say with REMOVE in the subject.
//
//   INPUT FILE
//
//     GetReply.ini: contains initialization parameters & filenames.
//
//   OUTPUT FILES (specified in GetReply.ini)
//
//     (1) Bounce.txt : list of email addresses from failure (bounced) emails from recipient's server.
//     (2) Remove.txt : list of email addresses from email addresses form "REMOVE" email replies.
//
//   BOUNCED EMAIL REPLIES
//
//   Email addresses corresponding to (previously sent) undeliverable email is written to the
//   'BounceFile' whose filename is read from GetReply.INI
//
//   In some case, mail returned as undeliverable can be due to a temporary problem with the recipient's
//   email server.
//
//   MATCHED EMAIL REPLIES
//
//   Email replies from recipients are processed. If a match is found in the subject ("REMOVE EmailAddress"),
//   then the email address is written to the 'RemoveFile" whose filename is read from GetReply.INI.
//
//   DELETE CODES
//
//   When reading email (server & recipient replies), MCM divides incoming email into three mutually
//   exclusive classes:
//
//   1. Failure messages ("bounced email") from the server.
//   2. User reply message (eg, "REMOVE email-address").
//   3. Everything else.
//
//   As each email is read when mcmReadReply is called, the 'DeleteCode' argument specifies if an email
//   is to be deleted after it is read.
//
//   DeleteCode    Action
//   ----------    ------
//        1        Delete if message is a failure message
//        2        Delete if message is a recipient reply message
//        4        Delete if message is not one of the above.
//
//   The above DeleteCodes can be added. For example, pass 3 in order to specify
//   that failure messages (DeleteCode = 1) and reply messages (DeleteCode = 2)
//   are to be deleted, but nothing else.
//
//   If no messages are to be deleted, use DeleteCode = 0
//
// STUNNEL
//
//   Stunnel is a free SSL proxy server that provides SSL services to Windows
//   programs. Stunnel is easy to install, very robust, and functions
//   transparently with very little overhead. Stunnel is required by MCM
//   application programs when connecting to an email server that requires SSL
//   services such as Gmail, Hotmail, Yahoo, and Microsoft Online.
//
//   For details on using Stunnel, see section 13 "Using Stunnel" in the MCM
//   User's Manual mcm_users.pdf in the DOCS directory or online at
//   http://www.marshallsoft.com/mcm_users.pdf
//
// ACCESS DELEGATION
//
//   Some web-based email servers (such as GMail) are requiring OAUTH access
//   delegation in addition to user authentication (such as TLS & SSL). See
//   http://www.marshallsoft.com/OAUTH_Access_Delegation.htm
//
// MANUALS: Can be found in \MCM4C\DOCS
//
//   Tutorial Manual (mcm_tutorial.pdf) Introduces the basic MCM functionality.
//   Servers Manual (mcm_servers.pdf) Background information on SMTP and POP3.
//   User Manual (mcm_users.pdf) Information that is not C/C++ specific.
//   Programmer Manual (mcm4c_programmer.pdf) Covers C/C++ specific information.
//   Reference Manual (mcm4c_reference.pdf) Lists individual MCM function.
//
// COMPILE: From the command line:
//
//     nmake -f GetReply64._M_  (64-bit Microsoft C/C++)
//
//     nmake -f GetReply32._M_  (32-bit Microsoft C/C++)
//      make -f GetReply32._B_  (Borland 5.0)
//      make -f GetReply32._I_  (Borland 5.5)
//      make -f GetReply32._G_  (GCC)
//      make -f GetReply32._L_  (LCC)
//     wmake -f GetReply32._W_  (Watcom 11)
//
//            GetReply$lcc.bat  (LCC-WIN32)
//            GetReply$gcc.bat  (MinGW GCC)
//
// VISUAL STUDIO: See vc_GetReply.cpp
//
//     (Links with Print.c and ReadINI.c)
//
// EXAMPLE: GetReply <history>
//
//   GetReply GetHistory.txt
//
//---------------------------------------------------------------------

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ReadINI.h"
#include "print.h"

#include "mcm.h"
#include "keycode.h"

//===============[ PROGRAMMER ]============================
//-- connect to SMTP that requires SSL
     //#define ENABLE_SSL
//===============[ END-PROGRAMMER ]========================

static FILE *fp = NULL; // history file pointer
static char *FileINI = "\\mcm4c\\apps\\GetReply.ini"; // email parameter file


//---[ SSL Proxy Data ]------------------------------------
// Please read \MCM4C\SSL\SSL_READ_ME.txt
#ifdef ENABLE_SSL
static int ProxyPort = 8802;  // proxy port for SSL (can use any local unused port)
static char *ProxyDir  = "\\MCM4C\\SSL";                // proxy directory (must be writable)
static char *ProxyCert = "\\MCM4C\\SSL\\STUNNEL.PEM";   // proxy certificate
static char *ProxyExe  = "\\MCM4C\\SSL\\STUNNEL.EXE";   // proxy executable
#endif

//---------------------------------------------------------

void ExitProgram(int ExitCode)
{Print("(exiting in 3 seconds)...\n");
 if(fp!=NULL) fclose(fp);
 mcmSleep(3000);
 mcmRelease();
 exit(ExitCode);
}

//---------------------------------------------------------
//
// Display error code then exit program
//

void ShowErrorAndExit(int ErrCode, char *Message)
{int Code;
 char Buffer[256];
 Print_I("MCM Error %d: ", ErrCode);
 if(Message)Print_S("(%s) \n", Message);
 // error ?
 if(ErrCode<0)
   {// get MCM error message
    Code = mcmGetError(ErrCode, (char *)Buffer, 255);
    if(Code) Print_S("%s\n", Buffer);
    Code = mcmGetInteger(MCM_GET_LIST_ERROR_STRING);
    // show string # if error was in recipient-list
    if(Code>0)Print_I("At recipient-list string # %d\n", Code);
    ExitProgram(1);
   }
}

//------------------------------------------------------------------------------
//
// read string from INI file

void ReadStringFromINI(char *ParamName, char *Buffer)
{int Code;
 // read paramter into 'Buffer'
 Code = iniRead(ParamName, Buffer);
 if(Code==0)
   {printf("Not expecting empty string for '%s'\n", ParamName);
    ExitProgram(2);
   }
 // check that user has edited RHS string
 if(Buffer[0]=='!')
   {printf("Edit string '%s' in file '%s'\n", Buffer, FileINI);
    ExitProgram(2);
   }
}

//------------------------------------------------------------------------------
//
// read integer from INI file
//

int ReadIntegerFromINI(char *ParamName)
{int Code;
 char Buffer[256];  // buffer for reading INI file parameters
 // read parameter into 'Buffer'
 Code = iniRead(ParamName, (char *)Buffer);
 if(Code==0)
   {Print_S("%s is missing integer value\n", ParamName);
    ExitProgram(2);
   }
 // convert to integer
 Code = atoi(Buffer);
 return Code;
}

//------------------------------------------------------------------------------
// add user-defined bounce string

void AddUserDefinedBounceString(char *String)
{Print_S("Adding bounce string '%s'\n", String);
 mcmSetString(MCM_ADD_BOUNCE_STRING, String);
}

//------------------------------------------------------------------------------
// display all user defined bounce strings

void DisplayBounceStrings(void)
{int i;
 int Code;
 char Buffer[64];
 // display all user defined bounce string (1,2,...)
 for(i=1;;i++)
   {Code = mcmGetString2(MCM_GET_BOUNCE_STRING, i, (char *)Buffer, 64);
    if(Code<=0) break;
    Print_IS("%d: '%s'\n",i,Buffer);
   }
}

//------------------------------------------------------------------------------

void main(int argc, char *argv[])
{//--- runtime parameters (read from GetReply.INI) ---
 char POP3_Server[256];      // SMTP server name or IP address.
 int  POP3_Port;             // SMTP port (normally 110 or 995 for SSL)
 char POP3_User[256];        // SMTP user name (SMTP Authentication only)
 char POP3_Pass[256];        // SMTP password (SMTP Authentication only)
 char MCM_Folder[256] = "."; // File folder for logs & history file
 char BounceFile[256];       // 1st skip list - previously bounced addresses
 char RemoveFile[256];       // 2nd skip list - addresses to be removed
 char *HistoryFile;          // history (output) file
 //--- other variables ---
 int  i;                     // loop index
 int  Len;                   // string length
 int  mcmCode;               // mcmAttach return code
 int  Debug = MCM_DEBUG_LOW; // debug level (defined in mcm.h)
 int  DeleteCode = 0;        // 1=delete_bounced; 2=delete_matched; 4=delete_others
 int  Code;                  // MCM return code (negative values are errors)
 char Temp[256];             // temporary buffer
 char Buffer[256];           // buffer for reading INI file parameters
 int  NbrMessages;           // number of messages on POP3 server
 int  NbrBounced = 0;        // # undeliverable (bounced) emails
 int  NbrRemoved = 0;        // # emails user replied to with 'REMOVE'
 int  NbrOther = 0;          // # emails marked as 'other'
 char SubjectBuffer[256];    // subject buffer for emails
 char YesNo;                 // = 'Y' (for yes) or 'N' (for no)
 int  ReplyCode;             // mcmReadReply() return code (-1 = EOF; 0 = OK to send;

 // begin

 Print("[This is a CONSOLE MODE program & is designed to run from a command window]\n");
 Print("GetReply Example Program\n");

 if(argc!=2)
   {printf("\nError: Expecting 1 command line argument, not %d\n\n", argc - 1);
    printf("Usage: GetReply <history>\n");
    printf("   Ex: GetReply GetHistory.txt\n");
    printf(" Note: Edit GetReply.ini before running.\n");
    printf("(exiting in 3 seconds)...\n");
    Sleep(3000);
    exit(1);
   }

 // get history file name (write access) from command line
 HistoryFile = argv[1];

 // open history output file

 fp = fopen((char *)HistoryFile, "w");
 if(fp==NULL)
   {printf("GetReply: Can not open '%s' for write\n", HistoryFile);
    mcmSleep(1500);
    exit(1);
   }

 // pass history file pointer to Print module
 PrintInit(fp);

 //***
 //*** Step 1: Read configuration parameters
 //***

 // open parameters file GetReply.INI
 if(!iniOpen(FileINI)) ExitProgram(1);

 // read POP3 server name or IP address (in dotted decimal notation)
 ReadStringFromINI((char *)"POP3_Server", (char *)POP3_Server);

 // read POP3 port
 POP3_Port = ReadIntegerFromINI((char *)"POP3_Port");

 // read POP3 user name
 Code = iniRead((char *)"POP3_User", (char *)POP3_User);
 if(Code==0) ExitProgram(2);

 // read POP3 password
 ReadStringFromINI((char *)"POP3_Pass", (char *)POP3_Pass);

 // open MCM folder
 ReadStringFromINI((char *)"MCM_Folder", (char *)MCM_Folder);

 Print_S("MCM Folder    : %s\n", MCM_Folder);
 Print_S("History File  : '%s'\n", HistoryFile);

 // read file name of 'bounce' file
 ReadStringFromINI((char *)"Bounce", (char *)BounceFile);
 Print_S("Bounce File   : %s\n", BounceFile);

 // read file name of 'remove' file
 ReadStringFromINI((char *)"Remove", (char *)RemoveFile);
 Print_S("Remove File   : %s\n", RemoveFile);

 // read "Delete Bounced" code
 ReadStringFromINI((char *)"DEL_Bounced", (char *)Buffer);
 if((Buffer[0]=='Y')||(Buffer[0]=='y'))
   {DeleteCode += MCM_DELETE_BOUNCED;
    YesNo = 'Y';
   }
 else YesNo = 'N';
 Print_C("Del Bounced   : %c (matched bounce string)\n", YesNo);

 // read "Delete Matched" code
 ReadStringFromINI((char *)"DEL_Matched", (char *)Buffer);
 if((Buffer[0]=='Y')||(Buffer[0]=='y'))
   {DeleteCode += MCM_DELETE_MATCHED;
    YesNo = 'Y';
   }
 else YesNo = 'N';
 Print_C("Del Matched   : %c (matched reply string)\n", YesNo);

 // read "Delete Other" code
 ReadStringFromINI((char *)"DEL_Other", (char *)Buffer);
 if((Buffer[0]=='Y')||(Buffer[0]=='y'))
   {DeleteCode += MCM_DELETE_OTHER;
    YesNo = 'Y';
   }
 else YesNo = 'N';
 Print_C("Del Other     : %c\n", YesNo);

 //***
 //*** Step 2: Attach MarshallSoft Client Mailer
 //***

 // attach MCM (must be 1st MCM function called)
#ifdef _WIN64
 printf("Attaching MCM64.DLL\n");
#else
 printf("Attaching MCM32.DLL\n");
#endif
 mcmCode = mcmAttach(MCM_KEY_CODE, 0, 1, Debug, (char *)MCM_Folder);
 Print_I("mcmAttach returns %d\n", mcmCode);

 if(mcmCode < 0)
   {// get MCM error message
    Print_I("MCM_KEY_CODE = %d (keycode.h)\n", MCM_KEY_CODE);
    ShowErrorAndExit(mcmCode, (char *)"Cannot attach MCM32.DLL");
    exit(1);
   }

 mcmWriteToLog((char *)"--- GetReply ---");
 // purchased version always returns 999 for mcmCode
 if(mcmCode == 999)
   {  Print("MCM Version   : Purchased\n");
    Print_I("Customer ID   : %d\n", mcmGetInteger(MCM_GET_CUSTOMER_ID));
    mcmGetString(MCM_GET_REGISTRATION, (char *)Temp, 51);
    Print_S("Registration  : %s\n", Temp);
   }
 else Print("MCM Version   : FREE\n");

 Code = mcmGetString(MCM_GET_VERSION, (char *)Temp, 40);
 if(Code>0) Print_S("Version ID    : %s\n", Temp);

 // display parameters
 Print_S("POP3 Server   : %s\n", POP3_Server);
 Print_I("POP3 Port     : %d\n", POP3_Port);
 Print_S("POP3 User     : %s\n", POP3_User);
 Print_S("POP3 Pass     : %s\n", POP3_Pass);

#ifdef ENABLE_SSL
 // doing SSL ?
 Print_I("Proxy Port    : %d\n", ProxyPort);
 Print_S("Proxy Dir     : %s\n", ProxyDir);
 Print_S("Proxy Cert    : %s\n", ProxyCert);
 Print_S("Proxy EXE     : %s\n", ProxyExe);
 Code = mcmSetProxySSL(0,1,ProxyDir,ProxyCert,ProxyExe,ProxyPort);
 if(Code<0) ShowErrorAndExit(Code, (char *)"Cannot enable proxy server");
#endif

 //***
 //*** Step 3: Open bounced email file
 //***

 Code = mcmOpenBounce((char *)BounceFile);
 if(Code<0) ShowErrorAndExit(Code, (char *)"Cannot open 'bounced' file");

 //***
 //*** Step 4: Open reply file (or files)
 //***

 // open file for replies on subject line "REMOVE email-address"
 Code = mcmOpenReply((char *)RemoveFile, (char *)"REMOVE");
 if(Code<0) ShowErrorAndExit(Code, (char *)"Cannot open 'remove' file");

#if 1
 // OPTIONAL: add two more bounce strings: this code MUST come after
 //           mcmOpenBounce() and mcmOpenReply()
 AddUserDefinedBounceString((char *)"No such user here");
 AddUserDefinedBounceString((char *)"Cet utilisateur n'est pas ici");
#endif

#if 0
 // OPTIONAL: display all user defined bounce strings
 DisplayBounceStrings();
#endif

 //***
 //*** Step 5: Connect to the POP3 server
 //***

 // connect to POP3 server
 Print("Connecting to POP3 server...\n");
 NbrMessages = mcmPop3Connect((char *)POP3_Server, POP3_Port, (char *)POP3_User, (char *)POP3_Pass);
 if(NbrMessages<0) ShowErrorAndExit(NbrMessages, (char *)"Cannot connect to POP3 server");
 Print_I("%d messages on POP3 server\n", NbrMessages);

 if(NbrMessages>0)
   {// have messages

    Print("Checking email replies...\n");
    // check email replies
    for(i=1;i<=NbrMessages;i++)
      {Print_I("---[ Reading message %d ]----------\n", i);

       //***
       //*** Step 6: Read next reply
       //***

       ReplyCode = mcmReadReply((char *)SubjectBuffer, 255, DeleteCode);
       if(ReplyCode<0) ShowErrorAndExit(ReplyCode, (char *)"mcmReadReply() fails");

       // ReplyCode Meaning
       // --------- -------
       //       0   No such email number.
       //  1..998   REPLY:  Email subject matched the 'reply' string, as specified by mcmOpenReply().
       //     999   BOUNCE: Email subject matched the 'bounce' string, as specified by mcmOpenBounce().
       //    1000   OTHER:  Email was neither failure nor reply messages

       Print_S("  %s\n", SubjectBuffer);

       if(ReplyCode==999)
         {// email subject matched the 'bounce' string (address will be written to BounceFile)
          NbrBounced++;
          Print("  Email was undeliverable !\n");
          Print("  Recipient: ");
          Len = mcmGetString(MCM_GET_RECIPIENT, (char *)Temp, 255);
          if(Len>0) Print_S("%s\n", Temp);
          else Print("? \n");
         }
       else if(ReplyCode==1000)
         {Print("  Email was neither a 'reply' nor 'failure' message \n");
          NbrOther++;
         }
       else if(ReplyCode>0)
         {// user replied with 'REMOVE' (address will be written to RemoveFile)
          NbrRemoved++;
          Print_I("  Email matched reply string %d\n", ReplyCode);
         }
      }
   }

 //***
 //*** Step 7: Close connection to the server
 //***

 mcmPop3Close();
 Print_I("%d emails read\n", NbrMessages);
 Print_I("%d emails were failure (BOUNCE) messages\n", NbrBounced);
 Print_I("%d emails were reply (REMOVE) messages\n", NbrRemoved);
 Print_I("%d emails were neither failure nor reply messages\n", NbrOther);
 Print_I("%d emails were deleted\n", mcmGetInteger(MCM_GET_NBR_EMAILS_DELETED));
 ExitProgram(0);
}
