//---------------------------------------------------------------------
//
// NAME: SendMail.c
//
// TYPE: Win32/Win64 console mode.
//
// DATE: May 10, 2021.
//
// PURPOSE
//
//   This example program uses MCM to send a customized email to everyone
//   on a list of recipients. Afterwards, the GetReply program can be run
//   to check for replies.
//
//   Also see the MCM tutorial (mcm_tutorial.pdf).
//
// EMAIL PROTECTION
//
//   If your security software scans outgoing email, you may need to disable
//   this feature since some security software does not handle multiple
//   connections correctly.
//
// EMAIL SECURITY
//
//   Even when connecting to a SMTP server using secure authentication such as
//   SSL/TLS, the contents of the email are not necessarily protected since the
//   email passes through multiple networks and servers before being received.
//
//   That is, regardless of what SMTP server you connect to, always assume that
//   the contents of email is public.
//
// INCLUDE FILE
//
//   The include file keycode.h contains the purchaser's key code value,
//   or 0 for the evaluation version of MCM.
//
//   Note that the key code value can be passed directly to mcmAttach,
//   in which case the include file is not needed.
//
// PARAMETER FILE
//
//   The text file SendMail.ini contains several runtime parameters:
//
//   SMTP_Server : SMTP server name or IP address
//   SMTP_Port   : SMTP port (normally 25, 587, or 465)
//   SMTP_User   : SMTP user name (SMTP Authentication only)
//   SMTP_Pass   : SMTP password (SMTP Authentication only)
//   SMTP_From   : Sender's email address
//   SMTP_ReOpen : # emails to send before closing & reopening the connection.
//   SMTP_Delay  : if > 0, # seconds to delay before reopening connection
//   Bounce      : List of recipients whose addresses previously bounced.
//   Remove      : List of recipients who have asked to be removed
//   Skip        : Additional addresses to skip
//
// INPUT FILES
//
//   Header file : Contains header info (To:, Subject:, .etc)
//   Letter file : Letter to send (text or HTML)
//   List file   : List of recipients text strings
//
// OUTPUT FILE
//
//   History file: Runtime history file
//
// STUNNEL
//
//   Stunnel is a free SSL proxy server that provides SSL services to Windows
//   programs. Stunnel is easy to install, very robust, and functions
//   transparently with very little overhead. Stunnel is required by MCM
//   application programs when connecting to an email server that requires SSL
//   services such as Gmail, Hotmail, Yahoo, and Microsoft Online.
//
//   For details on using Stunnel, see section 13 "Using Stunnel" in the MCM
//   User's Manual mcm_users.pdf in the DOCS directory or online at
//   http://www.marshallsoft.com/mcm_users.pdf
//
// ACCESS DELEGATION
//
//   Some web-based email servers (such as GMail) are requiring OAUTH access
//   delegation in addition to user authentication (such as TLS & SSL). See
//   http://www.marshallsoft.com/OAUTH_Access_Delegation.htm
//
// MANUALS: Can be found in \MCM4C\DOCS
//
//   Tutorial Manual (mcm_tutorial.pdf) Introduces the basic MCM functionality.
//   Servers Manual (mcm_servers.pdf) Background information on SMTP and POP3.
//   User Manual (mcm_users.pdf) Information that is not C/C++ specific.
//   Programmer Manual (mcm4c_programmer.pdf) Covers C/C++ specific information.
//   Reference Manual (mcm4c_reference.pdf) Lists individual MCM function.
//
// COMPILE: From the command line:
//
//     nmake -f SendMail64._M_  (64-bit Microsoft C/C++)
//
//     nmake -f SendMail32._M_  (32-bit Microsoft C/C++)
//      make -f SendMail32._B_  (Borland 5.0)
//      make -f SendMail32._I_  (Borland 5.5)
//      make -f SendMail32._G_  (GCC)
//      make -f SendMail32._L_  (LCC)
//     wmake -f SendMail32._W_  (Watcom 11)
//
//            SendMail$lcc.bat  (LCC-WIN32)
//            SendMail$gcc.bat  (MinGW GCC)
//
// VISUAL STUDIO: See vc_SendMail.cpp
//
//   (Links with Print.c and ReadINI.c)
//
// EXAMPLES: SendMail <chans> <header> <letter> <list> <history>
//
//   SendMail 4 Letter.hdr Letter.htm List.txt SendHistory.txt
//   SendMail 1 Letter(meeting).hdr Letter(meeting).txt List(meeting).txt Sent.txt
//   SendMail 1 Letter(dentist).hdr Letter(dentist).htm List(dentist).txt Sent.txt
//   SendMail 1 Letter(widgets).hdr Letter(widgets).htm List(widgets).txt Sent.txt
//
//---------------------------------------------------------------------

#define _CRT_SECURE_NO_WARNINGS

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ReadINI.h"
#include "print.h"
#include "mcm.h"
#include "keycode.h"

//===============[ PROGRAMMER ]============================
//-- define MERGE_EMAIL_ONLY if email is to be merged but NOT sent.
     //#define MERGE_EMAIL_ONLY
//---------------------------------------------------------
//-- allow duplicate email addresses in recipient list (does not affect skip lists)
     //#define ALLOW_DUPLICATE_ADDRESSES
//---------------------------------------------------------
//-- allow empty fields in recipient list (except 1st field)
     //#define ALLOW_EMPTY_FIELDS
//---------------------------------------------------------
//-- change letter macro delimiter from ` to %
     //#define USE_PERCENT_IN_LETTER
//---------------------------------------------------------
//-- connect to SMTP that requires SSL
     //#define ENABLE_SSL
//---------------------------------------------------------
//-- ignore email recipients rejected by the server
//   //#define IGNORE_REJECTED_EMAIL
//---------------------------------------------------------
//-- specify trace of recipients & server's responses
//#define SET_RCPT_TRACE_FILE
//===============[ END-PROGRAMMER ]========================

//---[ SSL Proxy Data ]------------------------------------
// Please read \MCM4C\SSL\SSL_READ_ME.txt
#ifdef ENABLE_SSL
static int ProxyPort = 8801;  // proxy port for SSL (can use any local unused port)
static char *ProxyDir  = "\\MCM4C\\SSL";                // proxy directory (must be writable)
static char *ProxyCert = "\\MCM4C\\SSL\\STUNNEL.PEM";   // proxy certificate
static char *ProxyExe  = "\\MCM4C\\SSL\\STUNNEL.EXE";   // proxy executable
static int  ProxyFlag = 1;                                // 1 = display stunnel icon on taskbar, 0 = don't
#endif

//---[ globals ]-------------------------------------------

static FILE *fp = NULL;             // history file pointer
static int  Debug = MCM_DEBUG_LOW;  // debug level (defined in mcm.h)
static char *MCM_Folder = "\\mcm4c\\apps";               // MCM file folder for logs & history file
static char *FileINI    = "\\mcm4c\\apps\\SendMail.ini"; // email parameter file

#ifdef SET_RCPT_TRACE_FILE
// specify recipient trace file (w/o extension)
static char *TraceFile = "\\mcm4c\\apps\\rcpt_trace";
#endif

//---------------------------------------------------------

static void WriteStatistics(void)
{int Recipients;
 Recipients = mcmStatistics(MCM_STAT_TOTAL_RECIPIENTS);
 if(Recipients>0)
   {Print("*** Runtime Statistics ***\n");
    Print_I("Total recipients     : %d\n", Recipients);
    Print_I("Bracketed addresses  : %d\n", mcmStatistics(MCM_STAT_BRACKETED_ADDRESSES) );
    Print_I("Invalid addresses    : %d\n", mcmStatistics(MCM_STAT_INVALID_ADDRESSES) );
    Print_I("Duplicate addresses  : %d\n", mcmStatistics(MCM_STAT_DUPLICATE_ADDRESSES) );
    Print_I("Bad attachments      : %d\n", mcmStatistics(MCM_STAT_WITH_BAD_ATTACHMENT) );
    Print_I("Skipped addresses    : %d\n", mcmStatistics(MCM_STAT_SKIPPED_ADDRESSES) );
    Print_I("Unknown char sets    : %d\n", mcmStatistics(MCM_STAT_WITH_UNKNOWN_CHARSETS) );
    Print_I("Empty macro strings  : %d\n", mcmStatistics(MCM_STAT_WITH_EMPTY_MACRO_STRING) );
    Print_I("Empty list strings   : %d\n", mcmStatistics(MCM_STAT_WITH_EMPTY_LIST_STRING) );
    Print_I("Average connect time : %d mSec\n", mcmStatistics(MCM_STAT_AVG_CONNECT_TIME));
    Print_I("Average send time    : %d mSec\n", mcmStatistics(MCM_STAT_AVG_SEND_TIME));
   }
}

//---------------------------------------------------------

void ExitProgram(int ExitCode)
{Print("(exiting in 3 seconds) ...\n");
 WriteStatistics();
 Sleep(100);
 if(fp!=NULL) fclose(fp);
 Sleep(2500);
 printf("Releasing MCM32.DLL ...");
 mcmRelease();
 printf("\n");
 Sleep(500);
 _exit(ExitCode);
}

//---------------------------------------------------------
//
// Display error code then exit program
//

void ShowErrorAndExit(int ErrCode, char *Message)
{int Code;
 char Buffer[256];
 Print_I("MCM Error %d: ", ErrCode);
 if(Message)Print_S("(%s) \n", Message);
 // error ?
 if(ErrCode<0)
   {// get MCM error message
    Code = mcmGetError(ErrCode, (char *)Buffer, 255);
    if(Code) Print_S("%s\n", Buffer);
    Code = mcmGetInteger(MCM_GET_LIST_ERROR_STRING);
    // show string # if error was in recipient-list
    if(Code>0)Print_I("At recipient-list string # %d\n", Code);
    if(Debug>MCM_DEBUG_OFF) printf("Diagnostic log files written to '%s'\n", MCM_Folder);
    ExitProgram(1);
   }
}

//------------------------------------------------------------------------------
//
// read string from INI file

int ReadStringFromINI(char *ParamName, char *Buffer, int CanBeEmptyFlag)
{int Code;
 // read paramter into 'Buffer'
 Code = iniRead(ParamName, Buffer);
 if(Code==0)
   {if(!CanBeEmptyFlag)
      {printf("Not expecting empty string for '%s'\n", ParamName);
       ExitProgram(2);
      }
    // allow empty string
    Buffer[0] = '\0';
    return 0;
   }
 // check that user has edited RHS string
 if(Buffer[0]=='!')
   {printf("Edit string '%s' in file '%s'\n", Buffer, FileINI);
    ExitProgram(2);
   }
 //if(Code==0) printf("%s = NULL\n", ParamName);
 //else printf("%s = '%s'\n", ParamName, Buffer);
 return Code;
}

//------------------------------------------------------------------------------
//
// read integer from INI file
//

int ReadIntegerFromINI(char *ParamName)
{int Code;
 char Buffer[256];  // buffer for reading INI file parameters
 // read parameter into 'Buffer'
 Code = iniRead(ParamName, (char *)Buffer);
 if(Code==0)
   {Print_S("%s is missing integer value\n", ParamName);
    ExitProgram(2);
   }
 // convert to integer
 Code = atoi(Buffer);
 return Code;
}

//------------------------------------------------------------------------------

int DoesFileExist(char *FilePtr)
{int Code;
 Code = mcmFileInfo(MCM_DOES_FILE_EXIST, FilePtr);
 if(!Code) Print_S("Error: File '%s' does not exist\n", FilePtr);
 return Code;
}

//------------------------------------------------------------------------------

void main(int argc, char *argv[])
{//--- runtime parameters (read from the command line) ---
 int  MaxChannels;           // maximum # channels to use
 char *HeaderFile;           // letter header file
 char *LetterFile;           // letter to send (containing macros)
 char *ListFile;             // list of recipients
 char *HistoryFile;          // history (output) file
 //--- runtime parameters (read from file SendMail.INI) ---
 char SMTP_Server[256];      // SMTP server name or IP address.
 int  SMTP_Port;             // SMTP port
 char SMTP_User[256] = "\0"; // SMTP user name (SMTP Authentication only)
 char SMTP_Pass[256] = "\0"; // SMTP password (SMTP Authentication only)
 char SMTP_From[256];        // Sender's email address
 int  SMTP_ReOpen = 0;       // if > 0, # emails after which connection is closed & reopened
 int  SMTP_Delay = 0;        // # seconds to wait before reopening connection
 char BounceFile[256];       // 1st skip list - previously bounced addresses
 char RemoveFile[256];       // 2nd skip list - addresses to be removed
 char SkipFile[256];         // 3rd skip list - addresses to be removed
 //--- other variables ---
 int  i;
 int  Code;                  // MCM return code (negative values are errors)
 int  mcmCode;               // mcmAttach return code
 int  MergeCode;             // MCM code returned from mcmMergeNext()
 int  EmailsReady = 0;       // # emails ready to be sent
 int  SendRate;              // email send rate (emails/hour)

 char Temp[256];             // temporary buffer

 Print("[This is a CONSOLE MODE program & is designed to run from a command window]\n");
 Print("\nSendMail Example Program\n\n");
 Print("IMPORTANT: Security scanning of outgoing email should be disabled!\n\n");
#ifdef MERGE_EMAIL_ONLY
 Print ("(Merge Only)\n");
#endif

 if(argc!=6)
   {printf("Error: Expecting 5 command line arguments, not %d\n\n", argc - 1);
    printf("Usage: SendMail <chans> <header> <letter> <list> <history>\n");
    printf("Examples:\n");
    printf(" SendMail 4 Letter.hdr Letter.htm List.txt SendHistory.txt\n");
    printf(" SendMail 1 Letter(meeting).hdr Letter(meeting).txt List(meeting).txt Sent.txt\n");
    printf(" SendMail 1 Letter(dentist).hdr Letter(dentist).htm List(dentist).txt Sent.txt\n");
    printf(" SendMail 1 Letter(widgets).hdr Letter(widgets).htm List(widgets).txt Sent.txt\n");
    printf("Note: Edit SendMail.ini before running.\n");
    printf("(exiting in 3 seconds)...\n");
    Sleep(3000);
    ExitProgram(10);
   }

 // verify that files exists
 if(!DoesFileExist(argv[2])) ExitProgram(1);
 if(!DoesFileExist(argv[3])) ExitProgram(2);
 if(!DoesFileExist(argv[4])) ExitProgram(3);

 // read history file name (write access) from command line
 HistoryFile = argv[5];

 // open history output file
 fp = fopen((char *)HistoryFile, "w");
 if(fp==NULL)
   {printf("SendMail: Can not open '%s' for write\n", HistoryFile);
    ExitProgram(11);
   }

 // pass history file pointer to Print module
 PrintInit(fp);

 //***
 //*** Step 1: Read configuration parameters
 //***

 // get maximum # channels to use
 MaxChannels = atoi(argv[1]);
 if((MaxChannels<1)||(MaxChannels>32))
   {Print_I("Number of channels must be 1 to 32, not %d", MaxChannels);
    ExitProgram(21);
   }

 // get file name of letter to send from command line
 HeaderFile = argv[2];

 // get file name of letter to send from command line
 LetterFile = argv[3];

 // read file name of recipient list to send from command line
 ListFile = argv[4];

 // open parameters file SendMail.INI
 Print_S("Opening file '%s'\n", FileINI);
 if(!iniOpen(FileINI)) ExitProgram(1);

 // read SMTP server name or IP address (in dotted decimal notation)
 ReadStringFromINI((char *)"SMTP_Server", (char *)SMTP_Server, FALSE);

 // read SMTP port
 SMTP_Port = ReadIntegerFromINI((char *)"SMTP_Port");

 // read SMTP user name (OPTIONAL: SMTP Authentication only)
 ReadStringFromINI((char *)"SMTP_User", (char *)SMTP_User, TRUE);

 // read SMTP password (OPTIONAL: SMTP Authentication only)
 ReadStringFromINI((char *)"SMTP_Pass", (char *)SMTP_Pass, TRUE);

 // read sender's email address
 ReadStringFromINI((char *)"SMTP_From", (char *)SMTP_From, FALSE);
 if(SMTP_From[0] != '<')
   {Print_SS("Must enclose '%s' in brackets: Expecting '<%s>'\n",
      SMTP_From, SMTP_From);
    ExitProgram(12);
   }

 // read re-open email count (SMTP connection closed & reopened)
 SMTP_ReOpen = ReadIntegerFromINI((char *)"SMTP_ReOpen");

 // read re-open delay (SMTP connection closed & reopened)
 SMTP_Delay = ReadIntegerFromINI((char *)"SMTP_Delay");

 // read 'bounce' file name (addresses previously returned as undeliverable)
 ReadStringFromINI((char *)"Bounce", (char *)BounceFile, FALSE);

 // read 'remove' file name (addresses not to send to, even if in recipient list)
 ReadStringFromINI((char *)"Remove", (char *)RemoveFile, FALSE);

 // read 'skip' file name (OPTIONAL: addresses not to send to, even if in recipient list)
 ReadStringFromINI((char *)"Skip", (char *)SkipFile, TRUE);

 //***
 //*** Step 2: Attach MarshallSoft Client Mailer DLL
 //***

 // attach MCM (must be 1st MCM function called)
#ifdef _WIN64
 Print("Attaching MCM64.DLL\n");
#else
 Print("Attaching MCM32.DLL\n");
#endif
 mcmCode = mcmAttach(MCM_KEY_CODE, 3, MaxChannels, Debug, (char *)MCM_Folder);
 Print_I("mcmAttach returns %d\n", mcmCode);

 // check for error
 if(mcmCode < 0)
   {// get MCM error message
    Print_I("MCM_KEY_CODE = %d (keycode.h)\n", MCM_KEY_CODE);
    ShowErrorAndExit(mcmCode, (char *)"Cannot attach MCM32.DLL");
   }

 mcmWriteToLog((char *)"--- SendMail ---");
 // purchased version always returns 999 for mcmCode
 if(mcmCode == 999)
   {  Print("MCM Version   : Purchased\n");
    Print_I("Customer ID   : %d\n", mcmGetInteger(MCM_GET_CUSTOMER_ID));
    mcmGetString(MCM_GET_REGISTRATION, (char *)Temp, 51);
    Print_S("Registration  : %s\n", Temp);
   }
 else Print("MCM Version   : FREE \n");

 Print_S("History File  : '%s'\n", HistoryFile);
 Print_I("Debug Flag    : %d\n", Debug);

 Code = mcmGetString(MCM_GET_VERSION, (char *)Temp, 40);
 if(Code>0) Print_S("Version ID    : %s\n", Temp);

#ifdef ENABLE_SSL
 // doing SSL ?
 Print_I("Proxy Port    : %d\n", ProxyPort);
 Print_S("Proxy Dir     : %s\n", ProxyDir);
 Print_S("Proxy Cert    : %s\n", ProxyCert);
 Print_S("Proxy EXE     : %s\n", ProxyExe);
 Print_I("Proxy Flag    : %d\n", ProxyFlag);
 Code = mcmSetProxySSL(0,ProxyFlag,ProxyDir,ProxyCert,ProxyExe,ProxyPort);
 if(Code<0) ShowErrorAndExit(Code, (char *)"Cannot enable proxy server");
#else
 Print("(SSL is not enabled)\n");
#endif

 // display email parameters

 Print_S("SMTP Server   : %s\n", SMTP_Server);
 Print_I("SMTP Port     : %d\n", SMTP_Port);
 if(SMTP_User) if(SMTP_User[0]!='\0') Print_S("SMTP User     : %s\n", SMTP_User);
 if(SMTP_Pass) if(SMTP_Pass[0]!='\0') Print_S("SMTP Pass     : %s\n", SMTP_Pass);
 if(SMTP_ReOpen>0) Print_I("Re-open       : %d\n", SMTP_ReOpen);
 Print_I("Max Channels  : %d\n", mcmGetInteger(MCM_GET_MAX_CHANNELS) );

 Print_S("Email Header  : %s (letter header)\n", HeaderFile);
 Print_S("Email Letter  : %s (letter to send)\n", LetterFile);
 Print_S("Recipient List: %s (list of recipients)\n", ListFile);
 Print_S("Skip List #1  : %s (skip these recipients)\n", BounceFile);
 Print_S("Skip List #2  : %s (skip these recipients)\n", RemoveFile);
 if(SkipFile[0]) Print_S("Skip List #3  : %s (skip these recipients)\n", SkipFile);

 //***
 //*** Step 3: Set processing flags
 //***

#ifdef ALLOW_DUPLICATE_ADDRESSES
 // disable duplicate detection in recipient list (ie, allows duplicates)
 Code = mcmSetInteger(MCM_SET_DUPLICATE_DETECT, 0);
 if(Code<0) ShowErrorAndExit(Code, NULL);
 Print("Duplicate email addresses are allowed in recipient list\n");
#endif

#ifdef ALLOW_EMPTY_FIELDS
 // allow empty fields (except 1st field) in recipient list
 // eg: "TomSmith@comcast.net, ,Tom,Smith"
 Code = mcmSetInteger(MCM_ALLOW_EMPTY_FIELDS, 1);
 if(Code<0) ShowErrorAndExit(Code, NULL);
 Print("Empty fields are allowed in recipient list\n");
#endif

#ifdef USE_PERCENT_IN_LETTER
 // change letter delimiter from ` to %: letters must be written using % instead of `
 Code = mcmSetInteger(MCM_SET_MACRO_DELIMITER, '%'); // percent = 37. = 25H
 if(Code<0) ShowErrorAndExit(Code, NULL);
 Print("Changing letter macro delimiter from ` to %\n");
#endif

#ifdef IGNORE_REJECTED_EMAIL
 // rejected email is ignored (default)
 Print("Email recipients rejected by the server will be ignored\n");
#else
 Print("Email recipients rejected by the server will NOT be ignored\n");
 Code = mcmSetInteger(MCM_IGNORE_REJECTED, 0);
 if(Code<0) ShowErrorAndExit(Code, NULL);
#endif

#ifdef SET_RCPT_TRACE_FILE
 // set recipient trace file (w/o extension)
 Code = mcmSetString(MCM_SET_RCPT_TRACE_FILE, TraceFile);
 if(Code<0) ShowErrorAndExit(Code, NULL);
 Print_S("Trace file '%s*' is enabled\n", TraceFile);
#endif

 //***
 //*** Step 4: Open the list of recipients
 //***

 Print_S("Opening file '%s'\n", ListFile);
 Code = mcmOpenList((char *)ListFile);
 if(Code<0)
   {int LineNbr = mcmGetInteger(MCM_GET_LIST_LINE_NBR);
    Print_S("Error in list '%s'\n", ListFile);
    if(LineNbr>0) Print_I("(at line %d)\n", LineNbr);
    ShowErrorAndExit(Code, (char *)"Cannot load list (of recipients)");
   }
 else Print_SI("Recipient list '%s' loaded (%d lines)\n", ListFile, Code);

 //***
 //*** Step 5: Open the letter header (contains To:, Subject:, etc)
 //***

 Print_S("Opening file '%s'\n", HeaderFile);
 Code = mcmOpenHeader((char *)HeaderFile);
 if(Code<0)
    {Print_S("Error in header file '%s'\n",HeaderFile );
     ShowErrorAndExit(Code, (char *)"Cannot load header file");
    }
 else Print_SI("Email header file '%s' loaded (%d lines)\n", HeaderFile, Code);

 //***
 //*** Step 6: Open the letter to be sent
 //***

 Print_S("Opening file '%s'\n",LetterFile);
 Code = mcmOpenLetter((char *)LetterFile);
 if(Code<0)
   {int CharPos = mcmGetInteger(MCM_GET_LETTER_CHAR_POS);
    int LineNbr = mcmGetInteger(MCM_GET_LETTER_LINE_NBR);
    Print_S("Error in letter '%s'\n",LetterFile );
    if(LineNbr>0) Print_II("(at line %d, character position %d)\n", LineNbr, CharPos );
    ShowErrorAndExit(Code, (char *)"Cannot load letter");
   }
 else Print_SI("Email letter '%s' loaded (%d lines)\n", LetterFile, Code);
 if(Code<1) ExitProgram(1);

 //***
 //*** Step 7: Open skip lists
 //***

 // Specify the list of recipients (BounceFile) whose email was previously
 // returned as undeliverable ("bounced" email), even if the email address
 // appears in the list of recipients (ListFile).

 Print_S("Opening file '%s'\n", BounceFile);
 Code = mcmOpenSkip((char *)BounceFile);
 if(Code<0) ShowErrorAndExit(Code, (char *)"Cannot load 'bounced' skip file");
 else Print_SI("1st skip file '%s' loaded (%d lines)\n", BounceFile, Code);

 // Specify list of recipients to skip (removed or "opt-out" list)
 // even if they appear in the list of recipients (ListFile)

 Print_S("Opening file '%s'\n", RemoveFile);
 Code = mcmOpenSkip((char *)RemoveFile);
 if(Code<0) ShowErrorAndExit(Code, (char *)"Cannot load 'removed' list (of recipients)");
 else Print_SI("2nd skip file '%s' loaded (%d lines)\n", RemoveFile, Code);

 // Specify another list of recipients to skip
 // even if they appear in the list of recipients (ListFile)

 if(SkipFile[0])
   {Print_S("Opening file '%s'\n", SkipFile);
    Code = mcmOpenSkip((char *)SkipFile);
    if(Code<0) ShowErrorAndExit(Code, (char *)"Cannot load 'skip' list (of recipients)");
    else Print_SI("3rd skip file '%s' loaded (%d lines)\n", SkipFile, Code);
   }

 //--
 //-- additional skip files [mcmOpenSkip] can be opened here
 //--

//***
//*** Step 8: Connect to SMTP server
//***

#ifdef MERGE_EMAIL_ONLY
 // merge only; no need to connect to server
#else
 //----- uncomment next line if using Verizon SMTP server (smtp.verizon.net)
 // mcmSetInteger(MCM_SET_SMTP_PROTOCOL, SMTP_AUTHENTICATE_LOGIN);
 //-----
 // connect to the SMTP server
 Print_SI("Connecting to server '%s' on port %d\n", SMTP_Server, SMTP_Port);
 // get # channels allowed (2 channels in FREE version, 32 in advance version)
 Code = mcmGetInteger(MCM_GET_ALLOWED_CHANNELS);
 if(Code<MaxChannels) MaxChannels = Code;
 Print_I("Requesting %d channels, please wait...\n", MaxChannels);
 Code = mcmSmtpConnect((char *)SMTP_Server, SMTP_Port, (char *)SMTP_User,
                       (char *)SMTP_Pass, (char *)SMTP_From, SMTP_ReOpen, SMTP_Delay);
 if(Code<0) ShowErrorAndExit(Code, NULL);

 Print_I("Connected to SMTP server on %d channels\n", Code);
#endif

 // send letter to each recipient in the list of recipients (unless in skip list)
 for(i=1;;i++)
   {mcmSleep(1);

    //***
    //*** Step 9: Merge letter with next recipient
    //***

    MergeCode = mcmMergeNext();

    // end-of-file ?
    if(MergeCode==MCM_EOF) break;
    Print_I("---- Merging Email (Recipient %d) ----\n", i);

    // error ?
    if(MergeCode<0) ShowErrorAndExit(MergeCode, (char *)"mcmMerge fails");
    mcmGetString(MCM_GET_RECIPIENT,(char *)Temp, 256);
    Print_S("Recipient : '%s'\n", Temp);

    // send to this recipient only if MergeCode = 0
    if(MergeCode>0)
       {char Temp[256];
        //-------------------------------------------------------
        // Handle all merge code (defined in mcm.h)
        //   1. Recipient found in skip list (MergeCode = skip_list_number = 1,2,...)
        //   2. MCM_MERGE_INVALID_ADDRESS : Invalid email address
        //   3. MCM_MERGE_DUPLICATE_ADDRESS : Duplicate email address
        //   4. MCM_MERGE_BRACKETS_NOT_ALLOWED : '<' and '>' not allowed in email address
        //   5. MCM_MERGE_CANNOT_OPEN_ATTACH : Cannot open attachment
        //   6. MCM_MERGE_UNKNOWN_CHARSET : Unknown character set
        //   7. MCM_MERGE_EMPTY_LIST_FIELD : Recipient list field is empty
        //-------------------------------------------------------
        mcmMergeText(MergeCode, (char *)Temp, 255);
        Print_S("Skipping this recipient - %s\n", Temp);
        // continue with next recipient
        continue;
       }

#ifdef MERGE_EMAIL_ONLY
    // don't send email
    Print_I("Email # %d merged\n", ++EmailsReady);
#else

    //***
    //*** Step 10: Send the mail
    //***

    // send letter to this recipient (MergeCode==0)
    Code = mcmSendMail();
    if(Code<0)
       {int Len;
        char Buffer[256];
        Len = mcmGetString(MCM_GET_LAST_RESPONSE, (char *)Buffer, 256);
        if(Len>0) Print((char *)Buffer);
        ShowErrorAndExit(Code, (char *)"mcmSendMail fails");
       }
    // report to user
    Print_III("Email # %d queued (%d emails sent, %d errors)\n",
       ++EmailsReady, mcmGetInteger(MCM_GET_EMAIL_SENT_COUNT),
       mcmGetInteger(MCM_GET_EMAIL_ERROR_COUNT) );

    // monitor send rate
    SendRate = mcmGetInteger(MCM_GET_SEND_RATE);
    Print_I("Sending at %d emails/hour\n", SendRate);
    // Sleep if sending too fast
    // ...

#endif
   } //end-for(i=1;;i++)

#ifdef MERGE_EMAIL_ONLY
  Print("Merging is complete\n");
#else

  //***
  //*** Step 11: Wait for all channels to complete
  //***

  Print_I("Waiting for %d channels to finish\n", mcmGetInteger(MCM_GET_CHANNEL_STATUS));
  mcmSleep(0);
  {int n;
   while(1)
     {// get # channels running
      n = mcmGetInteger(MCM_GET_CHANNEL_STATUS);
      if(n<=0) break;
      //
      // can do other things here waiting for last channel to complete
      //
      Print_I("Waiting for %d channels to finish\n", n);
      mcmSleep(500);
     }
  }

  //***
  //*** Step 12: Close connection on all channels
  //***

  // email has been accepted by server; close connection to SMTP server
  mcmSmtpClose();
  Print_II("%d emails successfully sent, %d had errors\n",
       mcmGetInteger(MCM_GET_EMAIL_SENT_COUNT), mcmGetInteger(MCM_GET_EMAIL_ERROR_COUNT) );
#endif

  // all done
  WriteStatistics();
  Sleep(100);
  if(fp!=NULL) fclose(fp);
  printf("Processing Completed\n");
  ExitProgram(0);
}
