//---------------------------------------------------------------------
//
// NAME: TestMCM.c
//
// TYPE: Win32/Win64 console mode.
//
// DATE: May 10, 2021.
//
// PURPOSE:
//
//   This program displays the MCM version and build.
//
// INCLUDE FILE
//
//   Include file keycode.h contains the purchaser's key
//   code value, or 0 for the evaluation version of MCM.
//
//   Note that the key code value can be passed directly to
//   mcmAttach, in which case the include file is not needed.
//
// MANUALS: Can be found in \MCM4C\DOCS
//
//   Tutorial Manual (mcm_tutorial.pdf) Introduces the basic MCM functionality.
//   Servers Manual (mcm_servers.pdf) Background information on SMTP and POP3.
//   User Manual (mcm_users.pdf) Information that is not C/C++ specific.
//   Programmer Manual (mcm4c_programmer.pdf) Covers C/C++ specific information.
//
// COMPILE: From the command line:
//
//     nmake -f TestMCM64._M_  (64-bit Microsoft C/C++)
//
//     nmake -f TestMCM32._M_  (32-bit Microsoft C/C++)
//      make -f TestMCM32._B_  (Borland 5.0)
//      make -f TestMCM32._I_  (Borland 5.5)
//      make -f TestMCM32._G_  (GCC)
//      make -f TestMCM32._L_  (LCC)
//     wmake -f TestMCM32._W_  (Watcom 11)
//
//            TestMCM$lcc.bat  (LCC-WIN32)
//            TestMCM$gcc.bat  (MinGW GCC)
//
// VISUAL STUDIO: See vc_TestMCM.cpp
//
//---------------------------------------------------------------------

//#define _CRT_SECURE_NO_WARNINGS

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "mcm.h"
#include "keycode.h"

//---------------------------------------------------------------------

void ExitProgram(int ExitCode)
{printf("\n(exiting in 3 seconds) ...\n");
 Sleep(3000);
 _exit(ExitCode);
}

//------------------------------------------------------------------------------

void main(void)
{int  Code;                  // MCM function return code
 int  Debug = MCM_DEBUG_LOW; // debug level (defined in mcm.h)
 char *MCM_Folder = "\\mcm4c\\apps";  // MCM file folder for logs & history file
 char Temp[256];             // temporary buffer

 printf("[This is a CONSOLE MODE program and is designed to run from a command window]\n\n");
 printf("TestMCM: Displays MCM version & build\n\n");

 // attach MCM (must be 1st MCM function called)
#ifdef _WIN64
 printf("Attaching MCM64.DLL\n");
#else
 printf("Attaching MCM32.DLL\n");
#endif

 Code = mcmAttach(MCM_KEY_CODE, 3, 1, Debug, (char *)MCM_Folder);
 printf("mcmAttach returns %d\n", Code);

 // check for attach error
 if(Code < 0)
   {// get MCM error message
    printf("mcmAttach returns error %d\n", Code);
    Code = mcmGetError(Code, (char *)Temp, 255);
    if(Code) printf("%s\n", Temp);
    printf("MCM_KEY_CODE = %d\n", MCM_KEY_CODE);
    exit(1);
   }

 mcmWriteToLog((char *)"--- TestMCM ---");
 // purchased version always returns 999
 if(MCM_KEY_CODE > 0)
   {printf("*** Advanced Version ***\n");
    printf("Customer ID    : %d\n", mcmGetInteger(MCM_GET_CUSTOMER_ID));
    // show registration
    mcmGetString(MCM_GET_REGISTRATION, (char *)Temp, 51);
    printf("Registration   : %s\n", Temp);
   }
 else
   {printf("*** FREE Version ***\n");
   }

 // show version & build of this DLL
 Code =  mcmGetString(MCM_GET_VERSION, (char *)Temp, 64);
 if(Code>0) printf("MCM Version    : %s\n", Temp );

 // show SEE version
 Code =  mcmGetString(MCM_GET_SEE_VERSION, (char *)Temp, 64);
 if(Code>0) printf("SEE Version    : %s\n", Temp);

 // show MCM folder (where logs are written)
 printf("MCM Folder     : %s\n", MCM_Folder);

 printf("Max Channels   : %d\n", mcmGetInteger(MCM_GET_ALLOWED_CHANNELS) );
 printf("Max Recip List : %d\n", mcmGetInteger(MCM_GET_ALLOWED_LIST_SIZE) );

 // all done
#ifdef _WIN64
 printf("\nReleasing MCM64.DLL\n");
#else
 printf("\nReleasing MCM32.DLL\n");
#endif

 mcmRelease();
 ExitProgram(0);
}
